#pragma once

#include <vector>
#include <time.h>

#include "colour.h"


//#define CELLDEBUG

// watercolour display modes
#define WCDM_PIGMENT	0
#define WCDM_SOLUTION	1
#define WCDM_FLUID		2

class wcCell
{
public:
	wcCell(){

		fluidCapacity = 5;
		fluidInfusedCapacity = 10;

		fluidContent = fluidDX = fluidDY = fluidInfused = 0;
	}
	
	// startup properties
	float fluidCapacity; // amount of fluid the paper in this cell can hold on the surface



	//dynamic properties
	float fluidContent;		// amount of fluid the paper in this cell holds on the surface
	float fluidDX, fluidDY; // amount of change of fluid each time step
	colCMY iPigDX, iPigDY;
	float iFluidDX, iFluidDY;

	float fluidInfusedCapacity;
	float fluidInfused;
	colCMY pigmentInfused;
	
	colCMY initialColour;
	colCMY pigmentContent;

	colRGB displayColour; // temp to test
	
	inline void operator = (wcCell rhs)
	{
		fluidCapacity = rhs.fluidCapacity;
		fluidContent = rhs.fluidContent;
		fluidDX = rhs.fluidDX;
		fluidDY = rhs.fluidDY;
		initialColour = rhs.initialColour;
		pigmentContent = rhs.pigmentContent;
		displayColour = rhs.displayColour;
		pigmentInfused = rhs.pigmentInfused;
		fluidInfused = rhs.fluidInfused;
		fluidInfusedCapacity = rhs.fluidInfusedCapacity;
	}

	
};



class wcCA
{
public:
	wcCA(int _caSize = 100, int _dCellSize = 5) {
		
		mCellArray.resize(_caSize*_caSize);
		//absorbencyField.resize(_caSize*_caSize);
		heightField.resize(_caSize*_caSize);
		
		running = 0;
		caSize = _caSize;
		dCellSize = _dCellSize;
		displayMode = WCDM_PIGMENT;
		displayGrid = 0;

		//define constants
		cEvaporation		= 0.01;
		cSpreading			= 0.01;
		cSurfaceTension		= 0.002;
		cGravity			= 1;
		cAbsorptionRate		= 0;
		cAbsorbedColourScale = 0.9;

		cWeightCyan = cWeightMagenta = cWeightYellow = 0.001;
	}

	
	void clear()
	{
		running = 0;
		mCellArray.clear();
		mCellArray.resize(caSize*caSize);
	}

	void init()
	{
		for (int i = 0; i < caSize*caSize; i++)
		{
			absorbencyField.push_back((float)(Random(1, 8))/100.0f);
		}
	}

	void draw();

	void update();

	void touchCell(int x, int y, colCMY col, float fluid);

	inline float getFluidContent(int x, int y);
	inline float getFluidInfused(int x, int y);
	inline colCMY getPigmentInfused(int x, int y);


	std::vector<wcCell> mCellArray;		// the array to hold all the cells!
	std::vector<float> absorbencyField;
	std::vector<float> heightField;
	int running;						// is the simulation running?
	int caSize;							// number of cells = caSize^2
	int dCellSize;						// the display size (in pixels) of a cell
	int displayMode;					// the display mode: pigment/solution/fluid
	int displayGrid;

	// some constants for the simulation
	float cEvaporation;					// how much fluid to remove every time step
	float cSpreading;					// how much the fluid spreads due to gravity
	float cSurfaceTension;				// how strongly the fluid should resist spreading
	float cGravity;						// duh! 
	float cAbsorptionRate;				// global scale for absorbtion

	float cAbsorbedColourScale;			// how much to scale down the colour that has been absorbed

	float cWeightCyan;
	float cWeightMagenta;
	float cWeightYellow;

private:
	int neighbours(int x, int y);
};

void wcCA::draw()
{
	// draw the cell array 
	glBegin(GL_TRIANGLES);
	wcCell* curCell;
	colRGB col;
	float fl;
	colCMY tmp;
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			curCell = &mCellArray[y*caSize + x];
			if (curCell->pigmentContent.c == 1 &&
				curCell->pigmentContent.m == 1 &&
				curCell->pigmentContent.y == 1 &&
				curCell->fluidContent == 0 ) continue; // if there's nothing to draw
			//select display colour based on display mode
			switch (displayMode)
			{
			case WCDM_PIGMENT:
				tmp = curCell->pigmentInfused*cAbsorbedColourScale;
				tmp += curCell->pigmentContent;
				col.fromCMY(tmp);
                break;
			case WCDM_SOLUTION:
				col.fromCMY(curCell->pigmentContent / curCell->fluidContent);
				break;
			case WCDM_FLUID:
				if (curCell->fluidContent > 10) fl = 1; else fl = curCell->fluidContent / 10;
				col.r = 1 - fl;
				col.g = 1 - fl;
				col.b = 1;
				break;
			}
			glColor3f(col.r, col.g, col.b);
			//glColor3f(mCellArray[y*caSize + x].displayColour.r,mCellArray[y*caSize + x].displayColour.g,mCellArray[y*caSize + x].displayColour.b);

			glVertex2f(x*dCellSize, y*dCellSize); //bottom left
			glVertex2f(x*dCellSize + dCellSize, y*dCellSize); //bottom right
			glVertex2f(x*dCellSize, y*dCellSize + dCellSize); // top left
			
			glVertex2f(x*dCellSize + dCellSize, y*dCellSize); //bottom right
			glVertex2f(x*dCellSize + dCellSize, y*dCellSize + dCellSize); //top right
			glVertex2f(x*dCellSize, y*dCellSize + dCellSize); // top left

		}
	}
	glEnd();


	//draw a grid
	if (displayGrid)
	{
		if (running) glColor3f(0.8, 0.8, 1.0);
		else glColor3f(0.8, 0.8, 0.8);
		
		glBegin(GL_LINES);
		for (int i = 0; i <= caSize; i++)
		{
			//vertical
			glVertex2i(i*dCellSize, 0);
			glVertex2i(i*dCellSize, caSize*dCellSize);

			//horizontal
			glVertex2i(0, i*dCellSize);
			glVertex2i(caSize*dCellSize, i*dCellSize);
		}
		glEnd();
	}
}

void wcCA::touchCell(int x, int y, colCMY col, float fluid)
{
	//mCellArray[(y/dCellSize) * caSize + (x/dCellSize)].displayColour.fromCMY(col);
	if (x >= caSize * dCellSize || x < 0) return;
	if (y >= caSize * dCellSize || y < 0) return;
	wcCell* curCell = (wcCell*)&(mCellArray[(y/dCellSize) * caSize + (x/dCellSize)]);

	if (curCell->fluidContent + fluid > curCell->fluidCapacity) return;

	curCell->fluidContent += fluid;
	curCell->pigmentContent.c += col.c;
	curCell->pigmentContent.m += col.m;
	curCell->pigmentContent.y += col.y;

	if (curCell->fluidContent < 0) curCell->fluidContent = 0;
	if (curCell->pigmentContent.c < 0) curCell->pigmentContent.c = 0;
	if (curCell->pigmentContent.m < 0) curCell->pigmentContent.m = 0;
	if (curCell->pigmentContent.y < 0) curCell->pigmentContent.y = 0;
	
	//colCMY tmp;
	//tmp.c = curCell->pigmentContent.c / curCell->fluidContent;
	//tmp.m = curCell->pigmentContent.m / curCell->fluidContent;
	//tmp.y = curCell->pigmentContent.y / curCell->fluidContent;
}

inline float wcCA::getFluidContent(int x, int y)
{
	if (x >= caSize ||
		y >= caSize ||
		x < 0 ||
		y < 0) return	0;

	else return mCellArray[y*caSize +x].fluidContent;
}
inline float wcCA::getFluidInfused(int x, int y)
{
	if (x >= caSize ||
		y >= caSize ||
		x < 0 ||
		y < 0) return	0;

	else return mCellArray[y*caSize +x].fluidInfused;
}

inline colCMY wcCA::getPigmentInfused(int x, int y)
{
	if (x >= caSize ||
		y >= caSize ||
		x < 0 ||
		y < 0) return	0;

	else return mCellArray[y*caSize +x].pigmentInfused;
}

void wcCA::update()
{
	std::vector<wcCell> tmpArray;
	tmpArray = mCellArray;

	wcCell* oldCell = NULL;
	wcCell* newCell = NULL;

#ifdef CELLDEBUG
	//*********************************** DEBUG OPEN FILE
	freopen("db_sf.txt", "w", stdout);
#endif
	
	// first calculate the amount of water displaced from each cell
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			oldCell = &mCellArray[y*caSize + x];
			newCell = &tmpArray[y*caSize + x];
			

			if (oldCell->fluidContent > 0)
			{
				// first let's calculate the x and y displacement surface forces due to spreading & surface tension
				float dxST = 0, dyST = 0;

				// grab and attenuate the fluid content from neighbouring cells
				float neighbourFluidX = 0;
				float neighbourFluidY = 0;
				float nextdoorFluidX = 0;
				float nextdoorFluidY = 0;
				for (int i = 1; i <= 10; i++)
				{
					//calculate and check value of cell offset so we don't overrun
					//if (x+i >= caSize || x-i < 0) continue;
					//if (y+i >= caSize || y-i < 0) continue;

					//neighbourFluidX += mCellArray[y*caSize + x+i].fluidContent / i;
					//neighbourFluidY += mCellArray[(y+i)*caSize + x].fluidContent / i;
					//neighbourFluidX += mCellArray[y*caSize + x-i].fluidContent / i;
					//neighbourFluidY += mCellArray[(y-i)*caSize + x].fluidContent / i;

					neighbourFluidX += getFluidContent(x+i, y) / i;
					neighbourFluidY += getFluidContent(x, y+i) / i;
					neighbourFluidX -= getFluidContent(x-i, y) / i;
					neighbourFluidY -= getFluidContent(x, y-i) / i;

					if (i == 1) 
					{
						//nextdoorFluidX = mCellArray[y*caSize + x-i].fluidContent - mCellArray[y*caSize + x+i].fluidContent;
						//nextdoorFluidY = mCellArray[(y-i)*caSize + x].fluidContent - mCellArray[(y+i)*caSize + x].fluidContent;
						nextdoorFluidX = getFluidContent(x-i, y) - getFluidContent(x+i, y);
						nextdoorFluidY = getFluidContent(x, y-i) - getFluidContent(x, y+i);
					}
				}

				dxST = neighbourFluidX * cSurfaceTension;
				dyST = neighbourFluidY * cSurfaceTension;

				// now due to spreading;
				float dxSP = nextdoorFluidX * cSpreading;
				float dySP = nextdoorFluidY * cSpreading;

				oldCell->fluidDX = cGravity * oldCell->fluidContent * (dxST + dxSP);
				oldCell->fluidDY = cGravity * oldCell->fluidContent * (dyST + dySP);

				
#ifdef CELLDEBUG				
				printf("Cell %d\n-----------------------\n", y*caSize+x);
				printf("oldCell->fluidContent   %.4f\n", oldCell->fluidContent);
				printf("neighbourFluidX			%.4f\n", neighbourFluidX);
				printf("neighbourFluidY			%.4f\n", neighbourFluidY);
				printf("nextdoorFluidX			%.4f\n", nextdoorFluidX);
				printf("nextdoorFluidY			%.4f\n", nextdoorFluidY);
				printf("dxST                    %.4f\n", dxST);
				printf("dyST                    %.4f\n", dyST);
				printf("dxSP                    %.4f\n", dxSP);
				printf("dySP                    %.4f\n", dySP);
				printf("oldCell->fluidDX        %.4f\n", oldCell->fluidDX );
				printf("oldCell->fluidDY        %.4f\n\n", oldCell->fluidDY);
#endif
			}
			else oldCell->fluidDX = oldCell->fluidDY = 0;
		}
	}

	float fromLeft = 0, fromRight = 0, fromAbove = 0, fromBelow = 0;
	

	// put this in to make sure the simulation doesn't explode!
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			oldCell = &mCellArray[y*caSize + x];
			newCell = &tmpArray[y*caSize + x];

			fromLeft = fromRight = fromAbove = fromBelow = 0;

			int xp1, xm1, yp1, ym1;
			xp1 = y*caSize + x+1;
			xm1 = y*caSize + x-1;
			yp1 = (y+1)*caSize + x;
			ym1 = (y-1)*caSize + x;

			if (x+1 < caSize)
				if (mCellArray[xp1].fluidDX < 0) 
					fromRight = fabs(mCellArray[xp1].fluidDX);
				else
					
					
			if (x-1 >= 0)
				if (mCellArray[xm1].fluidDX > 0) 
					fromLeft = fabs(mCellArray[xm1].fluidDX);
		
			if (y+1 < caSize)
				if (mCellArray[yp1].fluidDY < 0) 
					fromAbove = fabs(mCellArray[yp1].fluidDY);
				
				
			if (y-1 > 0)
				if (mCellArray[ym1].fluidDY > 0) 
					fromBelow = fabs(mCellArray[ym1].fluidDY);
			
			// if there's too much coming in, scale it down!
			float newFluid = (oldCell->fluidContent - (fabs(oldCell->fluidDX) + fabs(oldCell->fluidDY)) + fromLeft + fromRight + fromBelow + fromAbove);
			if (newFluid > oldCell->fluidCapacity)
			{
				float scale = (newFluid - oldCell->fluidCapacity) / newFluid;
				if (x+1 < caSize)
					if (mCellArray[xp1].fluidDX < 0) 
						mCellArray[xp1].fluidDX *= scale;

				if (x-1 >= 0)
					if (mCellArray[xm1].fluidDX > 0) 
						mCellArray[xm1].fluidDX *= scale;

				if (y+1 < caSize)
					if (mCellArray[yp1].fluidDY < 0) 
						mCellArray[yp1].fluidDY *= scale;

				if (y-1 > 0)
					if (mCellArray[ym1].fluidDY > 0) 
						mCellArray[ym1].fluidDY *= scale;
				
			}
			
		}
	}
	colCMY pLeft, pRight, pAbove, pBelow;
	fromLeft = 0, fromRight = 0, fromAbove = 0, fromBelow = 0;
	// now we've calculated the displacements for all cells, go through and update the fluid contents
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			oldCell = &mCellArray[y*caSize + x];
			newCell = &tmpArray[y*caSize + x];
			
			// first get fluid and pigment displacement from neighbouring cells
			int xp1, xm1, yp1, ym1;
			xp1 = y*caSize + x+1;
			xm1 = y*caSize + x-1;
			yp1 = (y+1)*caSize + x;
			ym1 = (y-1)*caSize + x;
			if (x+1 < caSize)
				if (mCellArray[xp1].fluidDX < 0) 
				{
					fromRight = fabs(mCellArray[xp1].fluidDX);
					if (mCellArray[xp1].fluidContent == 0) 
						pRight = 0;
					else
						pRight = mCellArray[xp1].pigmentContent * (fromRight / mCellArray[xp1].fluidContent);
				}
				else 
				{
					pRight = fromRight = 0;
				}
			

			if (x-1 >= 0)
				if (mCellArray[xm1].fluidDX > 0) 
				{
					fromLeft = fabs(mCellArray[xm1].fluidDX);
					if (mCellArray[xm1].fluidContent == 0) 
						pLeft = 0;
					else
						pLeft = mCellArray[xm1].pigmentContent * (fromLeft / mCellArray[xm1].fluidContent);
				}
				else 
				{ 
					pLeft = fromLeft = 0;
				}

			if (y+1 < caSize)
				if (mCellArray[yp1].fluidDY < 0) 
				{
					fromAbove = fabs(mCellArray[yp1].fluidDY);
					if (mCellArray[yp1].fluidContent == 0) 
						pAbove = 0;
					else
						pAbove = mCellArray[yp1].pigmentContent * (fromAbove / mCellArray[yp1].fluidContent);
				}
				else 
				{
					pAbove = fromAbove = 0;
				}

			if (y-1 > 0)
				if (mCellArray[ym1].fluidDY > 0) 
				{
					fromBelow = fabs(mCellArray[ym1].fluidDY);
					if (mCellArray[ym1].fluidContent == 0) 
						pBelow = 0;
					else
						pBelow = mCellArray[ym1].pigmentContent * (fromBelow / mCellArray[ym1].fluidContent);
				}
				else pBelow = fromBelow = 0;

			// now add it all together to get the new fluid content
			//newCell->fluidContent = oldCell->fluidContent - (fabs(oldCell->fluidDX) + fabs(oldCell->fluidDY)) + (fromLeft + fromBelow) + fabs(fromRight+fromAbove);
			newCell->fluidContent = oldCell->fluidContent - (fabs(oldCell->fluidDX) + fabs(oldCell->fluidDY)) + fromLeft + fromRight + fromBelow + fromAbove;
			if (newCell->fluidContent > newCell->fluidCapacity) newCell->fluidContent = newCell->fluidCapacity;
			// now work out how much pigment is transferred
			// pigment should be transferred in proportion to the amount of water
			float fluidOutProportion = -1;
			if (oldCell->fluidContent > 0)
			{
				fluidOutProportion = (fabs(oldCell->fluidDX) + fabs(oldCell->fluidDY)) / oldCell->fluidContent;
				newCell->pigmentContent = oldCell->pigmentContent - (oldCell->pigmentContent * fluidOutProportion);
			}
			

			// now figure out how much pigment is transferred in from each cell
			newCell->pigmentContent += pAbove;
			newCell->pigmentContent += pBelow;
			newCell->pigmentContent += pLeft;
			newCell->pigmentContent += pRight;

#ifdef CELLDEBUG
			///******************************************************** DEBUG OUTPUT
			printf("Cell %d\n-----------------------\n", y*caSize+x);
			printf("oldCell->fluidContent:     %.4f\n", oldCell->fluidContent);
			printf("fromLeft:                  %.4f\n", fromLeft);
			printf("fromRight:                 %.4f\n", fromRight);
			printf("fromAbove:                 %.4f\n", fromAbove);
			printf("fromBelow:                 %.4f\n", fromBelow);
			printf("fluidOutProportion:        %.4f\n", fluidOutProportion);
			printf("newCell->fluidContent:     %.4f\n\n", newCell->fluidContent);
			//*/
#endif

			// now take away the evaporation
			newCell->fluidContent -= cEvaporation;
			if (newCell->fluidContent < 0) newCell->fluidContent = 0;
			
		}
	}
	
	float fluidAbsorbed;
	colCMY pigmentAbsorbed;
	// now see how much water and pigment is absorbed into the paper
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			oldCell = &mCellArray[y*caSize + x];
			newCell = &tmpArray[y*caSize + x];

			if (oldCell->fluidContent == 0) continue;

			fluidAbsorbed = cAbsorptionRate * absorbencyField[y*caSize + x] * oldCell->fluidContent;
			if (fluidAbsorbed > (oldCell->fluidInfusedCapacity - oldCell->fluidInfused))
				fluidAbsorbed = oldCell->fluidInfusedCapacity - oldCell->fluidInfused;

			newCell->fluidInfused += fluidAbsorbed;
			newCell->fluidContent -= fluidAbsorbed;

			pigmentAbsorbed = oldCell->pigmentContent * (fluidAbsorbed / oldCell->fluidContent);
			newCell->pigmentContent -= pigmentAbsorbed;
			newCell->pigmentInfused += pigmentAbsorbed;


		}

	}
	
	float fluidGX, fluidGY;
	
	colCMY pigGX, pigGY;
	

	// now do the displacement of fluid within the paper
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			oldCell = &mCellArray[y*caSize + x];
			newCell = &tmpArray[y*caSize + x];

			fluidGX = oldCell->fluidInfused - getFluidInfused(x-1, y);
			fluidGY = oldCell->fluidInfused - getFluidInfused(x, y-1);

			pigGX = oldCell->pigmentInfused - getPigmentInfused(x-1, y);
			pigGY = oldCell->pigmentInfused - getPigmentInfused(x, y-1);

			oldCell->iFluidDX = cGravity * absorbencyField[y*caSize + x] * fluidGX;
			oldCell->iFluidDY = cGravity * absorbencyField[y*caSize + x] * fluidGY;

			oldCell->iPigDX.c = cGravity * absorbencyField[y*caSize + x] * cWeightCyan    * pigGX.c;
			oldCell->iPigDX.m = cGravity * absorbencyField[y*caSize + x] * cWeightMagenta * pigGX.m;
			oldCell->iPigDX.y = cGravity * absorbencyField[y*caSize + x] * cWeightYellow  * pigGX.y;

			oldCell->iPigDY.c = cGravity * absorbencyField[y*caSize + x] * cWeightCyan    * pigGY.c;
			oldCell->iPigDY.m = cGravity * absorbencyField[y*caSize + x] * cWeightMagenta * pigGY.m;
			oldCell->iPigDY.y = cGravity * absorbencyField[y*caSize + x] * cWeightYellow  * pigGY.y;

		}
	}

	// and finally move the infused fluid and pigment
	fromLeft = 0, fromRight = 0, fromAbove = 0, fromBelow = 0;
	// now we've calculated the displacements for all cells, go through and update the fluid contents
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			oldCell = &mCellArray[y*caSize + x];
			newCell = &tmpArray[y*caSize + x];

			// first get fluid and pigment displacement from neighbouring cells
			int xp1, xm1, yp1, ym1;
			xp1 = y*caSize + x+1;
			xm1 = y*caSize + x-1;
			yp1 = (y+1)*caSize + x;
			ym1 = (y-1)*caSize + x;

			if (x+1 < caSize)
			{
				if (mCellArray[xp1].iFluidDX < 0) 
					fromRight = fabs(mCellArray[xp1].iFluidDX);	
				else
					fromRight = 0;

				if (mCellArray[xp1].iPigDX.c < 0)
				{
					pRight.c = fabs(mCellArray[xp1].iPigDX.c);
					pRight.m = fabs(mCellArray[xp1].iPigDX.m);
					pRight.y = fabs(mCellArray[xp1].iPigDX.y);
				}
				else
					pRight = 0;
			}
				
			if (x-1 >= 0)
			{
				if (mCellArray[xm1].iFluidDX > 0) 
					fromLeft = fabs(mCellArray[xm1].iFluidDX);
				else
					fromLeft = 0;

				if (mCellArray[xm1].iPigDX.c > 0)
				{
					pLeft.c = fabs(mCellArray[xm1].iPigDX.c);
					pLeft.m = fabs(mCellArray[xm1].iPigDX.m);
					pLeft.y = fabs(mCellArray[xm1].iPigDX.y);
				}
				else
					pLeft = 0;
			}
					

			if (y+1 < caSize)
			{
				if (mCellArray[yp1].iFluidDY < 0) 
					fromAbove = fabs(mCellArray[yp1].iFluidDY);
				else 
					fromAbove = 0;

				if (mCellArray[yp1].iPigDY.c < 0)
				{
					pAbove.c = fabs(mCellArray[yp1].iPigDY.c);
					pAbove.m = fabs(mCellArray[yp1].iPigDY.m);
					pAbove.y = fabs(mCellArray[yp1].iPigDY.y);
				}
				else
					pAbove = 0;
			}
	
			if (y-1 > 0)
			{
				if (mCellArray[ym1].iFluidDY > 0) 
					fromBelow = fabs(mCellArray[ym1].iFluidDY);
				else 
					fromBelow = 0;

				if (mCellArray[ym1].iPigDY.c > 0)
				{
					pBelow.c = fabs(mCellArray[ym1].iPigDY.c);
					pBelow.m = fabs(mCellArray[ym1].iPigDY.m);
					pBelow.y = fabs(mCellArray[ym1].iPigDY.y);
				}
				else
					pBelow = 0;
			}

			// now add it all together!
			newCell->fluidInfused -= (fabs(oldCell->iFluidDX) + fabs(oldCell->iFluidDY)) + fromAbove + fromBelow + fromLeft + fromRight;
			if (newCell->fluidInfused < 0) newCell->fluidInfused = 0;
			
			colCMY tmp;
			tmp.c = fabs(oldCell->iPigDX.c) + fabs(oldCell->iPigDY.c);
			tmp.m = fabs(oldCell->iPigDX.m) + fabs(oldCell->iPigDY.m);
			tmp.y = fabs(oldCell->iPigDX.y) + fabs(oldCell->iPigDY.y);
			newCell->pigmentInfused -= tmp + pAbove + pBelow + pLeft + pRight;
		}
	}

	mCellArray = tmpArray;
}