#pragma once


#include "matrix4.h"
#include <vector>

// define the behaviour identifiers
#define BID_SEEK			01200
#define BID_FLEE			01201
#define BID_PURSUIT			01202
#define BID_ARRIVE			01203
#define BID_OFFSETARRIVE	01204
#define BID_OFFSETPURSUIT	01205
#define BID_WANDER			01206
#define BID_AVOIDWALLS		01207
#define BID_UCA				01301
#define BID_SEPERATION		01302





// Helper class colour - for boid colours
class Colour 
{
public:
	Colour(){
		r = 0; g = 0; b = 0;
	}
	Colour(float _r, float _g, float _b) {
		r = _r; g = _g; b = _b;
	}

	inline void operator = (Colour c) {
		r = c.r; g = c.g; b = c.b;
	}

	void set (float _r, float _g, float _b) {
		r = _r; g = _g; b = _b;
	}

	float r, g, b;
};

// Helper class wall, so we can keep em inside it!
class WallSegment
{
public:
	WallSegment(Vec3 _start, Vec3 _end, Vec3 _normal){
		start =		_start;
		end =		_end;
		normal =	_normal;
	
	};
	~WallSegment(){};

	Vec3 start, end, normal;
};


class Wall
{
public:
	Wall(){};
	~Wall(){};
	
	void addSegment(WallSegment* ws) {
		segs.push_back(ws);

	}

	void addSegment(Vec3 start, Vec3 end, Vec3 normal) {
		WallSegment* tmp = new WallSegment(start, end, normal);
		segs.push_back(tmp);
	}

	inline void draw(void) {
		glColor3f(colour.r, colour.g, colour.b);
		
		glBegin(GL_LINES);
		for (unsigned i = 0; i < segs.size(); i++)
		{
			glVertex3f(segs[i]->start.x, segs[i]->start.y, segs[i]->start.z);
			glVertex3f(segs[i]->end.x, segs[i]->end.y, segs[i]->end.z);
		}
		glEnd();
		
		
	}

	std::vector<WallSegment*> segs;
	Colour colour;
};



// Abstract base class to allo us to use this in Behaviour, and Behaviour in Boid, which we will derive from this
class B {
public:
	B(){}
	~B(){}

	// Some helper functions for converting points to and from local co-ordinates
	
	inline Vec3 toWorldv(Vec3 v){

		Mat4 t;
		t._00 = side.x;	t._10 = side.y;	t._20 = side.z;	t._30 = -(position*side);
		t._01 = up.x;		t._11 = up.y;		t._21 = up.z;		t._31 = -(position*up);
		t._02 = heading.x;		t._12 = heading.y;		t._22 = heading.z;		t._32 = -(position*heading);

		t = t.inverse();

		return t * v;

	}

	inline Vec3 toLocalv(Vec3 v) {

		Mat4 t;
		t._00 = side.x;	t._10 = side.y;	t._20 = side.z;	t._30 = -(position*side);
		t._01 = up.x;		t._11 = up.y;		t._21 = up.z;		t._31 = -(position*up);
		t._02 = heading.x;		t._12 = heading.y;		t._22 = heading.z;		t._32 = -(position*heading);

		return t * v;

	}

	inline Vec3 toWorldp(Vec3 v){

		Mat4 t;
		t._00 = side.x;	t._10 = side.y;	t._20 = side.z;	t._30 = -(position*side);
		t._01 = up.x;		t._11 = up.y;		t._21 = up.z;		t._31 = -(position*up);
		t._02 = heading.x;		t._12 = heading.y;		t._22 = heading.z;		t._32 = -(position*heading);

		t = t.inverse();

		return t.multAsPoint(v);

	}

	inline Vec3 toLocalp(Vec3 v)  {

		Mat4 t;
		t._00 = side.x;	t._10 = side.y;	t._20 = side.z;	t._30 = -(position*side);
		t._01 = up.x;		t._11 = up.y;		t._21 = up.z;		t._31 = -(position*up);
		t._02 = heading.x;		t._12 = heading.y;		t._22 = heading.z;		t._32 = -(position*heading);

		return t.multAsPoint(v);

	}

    // Makes this an abstract class
	virtual void draw(int ) = 0;


	Vec3 position;
	Vec3 heading;	//-z
	Vec3 side;		//+x
	Vec3 up;		//+y

	Colour colour;

	float maxSpeed;	// units per second
	float maxForce;
	float mass;
	Vec3 velocity;

	Vec3 top;
	Vec3 left;
	Vec3 right;

};


// abstract base class for inheritance
class Behaviour {
public:
	Behaviour(){};
	~Behaviour(){};
	virtual Vec3 calculate()=0;
	
	// how much does this behaviour affect the boid's steering?
	float weight;
	int bId;
};

// Seek behaviour
class BSeek : public Behaviour
{
public:
	BSeek(B* _thisBoid, Vec3& _targetPos, float _weight) {
		thisBoid = _thisBoid;
		targetPos = _targetPos;
		seekingBoid = false;
		weight = _weight;
		bId = BID_SEEK;
	}
	BSeek(B* _thisBoid, B* _targetBoid, float _weight) {
		thisBoid = _thisBoid;
		targetBoid = _targetBoid;
		seekingBoid = true;
		weight = _weight;
		bId = BID_SEEK;
	}
	~BSeek(){}

	void setTarget(Vec3& _targetPos) {
		targetPos = _targetPos;
		seekingBoid = false;
	}

	void setTarget(B* _targetBoid) {
		targetBoid = _targetBoid;
		seekingBoid = true;
	}

	Vec3 calculate() {
		if (seekingBoid && targetBoid)
		{
			return (targetBoid->position - thisBoid->position) - thisBoid->velocity;
		}
		//implied else seeking target position
		return targetPos - thisBoid->position;
	}

	Vec3 targetPos;
	B* targetBoid;
	B* thisBoid;

	bool seekingBoid;
};

// Arrival behaviour
class BArrive : public Behaviour
{
public:
	~BArrive(){}
	BArrive(B* _thisBoid, B* _targetBoid, float _weight)
	{
		seekingBoid = true;
		targetBoid = _targetBoid;
		thisBoid = _thisBoid;
		weight = _weight;
		bId = BID_ARRIVE;

	}
	BArrive(B* _thisBoid, Vec3 _targetPos, float _weight)
	{
		seekingBoid = false;
		targetPos = _targetPos;
		thisBoid = _thisBoid;
		weight = _weight;
		bId = BID_ARRIVE;
	}

	Vec3 calculate()
	{
		if (seekingBoid)
		{
			const float decelTime = 20.0f;
			Vec3 targetOffset = targetBoid->position - thisBoid->position;
			
			//if (targetOffset.length() < 0.1) return Vec3(0, 0, 0);			

			float speed = targetOffset.length() / decelTime;
			float clipSpeed;
			
			if (speed <= thisBoid->maxSpeed) clipSpeed = speed;
			else clipSpeed = thisBoid->maxSpeed;
			
			return  (targetOffset * (clipSpeed / targetOffset.length())) - thisBoid->velocity;
		}
		
	}
	
	Vec3 targetPos;
	B* targetBoid;
	B* thisBoid;

	bool seekingBoid;
	
};


class BPursuit : public Behaviour
{
public:
	BPursuit(B* _thisBoid, B* _targetBoid, float _weight){
		thisBoid = _thisBoid;
		targetBoid = _targetBoid;
		weight = _weight;
		bId = BID_PURSUIT;
	}
	~BPursuit(){}

	Vec3 calculate()
	{
		Vec3 targetOffset = targetBoid->position - thisBoid->position;
		float dist = targetOffset.length();

		// predict the target's future position
		Vec3 targetFuturePos = targetBoid->position +  targetBoid->velocity * (dist/thisBoid->maxSpeed);
		Vec3 desiredVelocity = (targetFuturePos - thisBoid->position).normal() * thisBoid->maxSpeed;
		return desiredVelocity - thisBoid->velocity;
	}



	B* targetBoid;
	B* thisBoid;

};


class BOffsetPursuit : public Behaviour
{
public:
	BOffsetPursuit(B* _thisBoid, B* _targetBoid, Vec3 _offset, float _weight){
		thisBoid = _thisBoid;
		targetBoid = _targetBoid;
		offset = _offset;
		weight = _weight;
		bId = BID_OFFSETPURSUIT;
	}
	~BOffsetPursuit(){}

	Vec3 calculate()
	{
		Vec3 targetOffset = (targetBoid->position + targetBoid->toWorldp(offset)) - thisBoid->position;
		float dist = targetOffset.length();

		// predict the target's future position
		Vec3 targetFuturePos = (targetBoid->position + targetBoid->toWorldp(offset)) +  targetBoid->velocity * (dist/thisBoid->maxSpeed);
		Vec3 desiredVelocity = (targetFuturePos - thisBoid->position).normal() * thisBoid->maxSpeed;
		return desiredVelocity - thisBoid->velocity;
	}



	B* targetBoid;
	B* thisBoid;
	Vec3 offset;

};

// Arrival to an offset behaviour
class BOffsetArrive : public Behaviour
{
public:
	~BOffsetArrive(){}
	BOffsetArrive(B* _thisBoid, B* _targetBoid, Vec3 _offset, float _weight)
	{

		targetBoid = _targetBoid;
		thisBoid = _thisBoid;
		offset = _offset;
		weight = _weight;
		bId = BID_OFFSETARRIVE;

	}
	
	Vec3 calculate()
	{

			const float decelTime = 20.0f;
			Vec3 targetOffset = targetBoid->toWorldp(offset) - thisBoid->position;

			//if (targetOffset.length() < 0.1) return Vec3(0, 0, 0);			

			float speed = targetOffset.length() / decelTime;
			float clipSpeed;

			if (speed <= thisBoid->maxSpeed) clipSpeed = speed;
			else clipSpeed = thisBoid->maxSpeed;

			return  (targetOffset * (clipSpeed / targetOffset.length())) - thisBoid->velocity;


	}

	B* targetBoid;
	B* thisBoid;

	Vec3 offset;

};


class BWander : public Behaviour
{
public:
	BWander(B* _thisBoid, Vec3 _circleOffset, float _circleRadius, float _weight)
	{
		thisBoid = _thisBoid;
		circleOffset = _circleOffset;
		circleRadius = _circleRadius;
		weight = _weight;
		first = true;
		bId = BID_WANDER;
	}
	~BWander(){};

	Vec3 calculate()
	{
		static Vec3 aim;
		if (first)
		{
			// start off with a radius vector (0, 0, 1), rotate it around y by random degrees
			Vec3 r(0, 0, 1);
			Mat4 rot;
			rot = rot.rotateDeg(rand()%360, Y);
			// our 'aim' vector will be this new r + circleOffset
			aim = r + circleOffset;
			first = false;
		}

		Vec3 p = aim + Vec3((float)((rand()%1000)/1000.0f)-0.5f, 0, (float)((rand()%1000)/1000.0f)-0.5f);
		Vec3 cp = -(circleOffset - p);
		cp.truncate(circleRadius);
		aim = circleOffset + cp;
		return thisBoid->toWorldv(aim) - thisBoid->velocity;
		

	}

	Vec3 circleOffset;
	float circleRadius;

	B* thisBoid;
	bool first;
};

// return the distance along the given normalized ray to the given plane
inline float distRayPlane(Vec3 rayO, Vec3 rayV, Vec3 planeNormal, float planeD)
{
	float cosAlpha = rayV * planeNormal;
	float deltaD = planeD - rayO * planeNormal;
	
	// don't divide by zero!
	if (cosAlpha == 0) return -1.0f;

	return deltaD/cosAlpha;
}

class BAvoidWalls : public Behaviour
{
public:
	BAvoidWalls(Wall* _wall, float _activationDistance, B* _thisBoid, float _weight) {
		thisBoid = _thisBoid;
		activationDistance = _activationDistance;
		weight = _weight;
		wall = _wall;
		bId = BID_AVOIDWALLS;
	}
	~BAvoidWalls(){};

	Vec3 calculate()
	{
		WallSegment* closestSeg;
		float closestDist = 99999999999.0f;
		Vec3 rayO = thisBoid->position;
		Vec3 rayV = thisBoid->heading;
		// loop through and find the closest wall segment
		for (unsigned i = 0; i < wall->segs.size(); i++)
		{
			// find out how far this wall segment is from us
			float D = wall->segs[i]->normal * wall->segs[i]->start;
			float thisDist = (D- (wall->segs[i]->normal * rayO)) / (wall->segs[i]->normal * rayV);
			if (thisDist < closestDist && thisDist > 0) // if it's the closest
			{
				closestDist = thisDist;
				closestSeg = wall->segs[i];
			}
		}

		// now take the appropriate action!
		// check that we need to do anything at all
		if (closestDist > activationDistance) return Vec3(0, 0, 0);

		Vec3 intersectionPoint = thisBoid->position + thisBoid->heading * closestDist;

		// just adding the normal scaled up by a lot will keep them going through the corners
		return (intersectionPoint + (closestSeg->normal*5) - thisBoid->position) - thisBoid->velocity;
		

	}

	B* thisBoid;
	Wall* wall;
	float activationDistance;

};

// some definitons for our closest approach function
#define NO_COLLISION	04300
//#define CLOSE_APPROACH	04301
#define COLLISION		04302
#define EPS				0.0001


class BUCA : public Behaviour
{
public:
	BUCA(std::vector<B*>* _allBoids, B* _thisBoid, float _weight){
		allBoids = _allBoids;
		thisBoid = _thisBoid;
		weight = _weight;
		bId = BID_UCA;
	};
	~BUCA(){};
	
	int findClosestApproach(B* vehicle1, B* vehicle2, float maxTime, float& time, float& distanceSqr)
	{
		Vec3 p = vehicle2->position - vehicle1->position;
		Vec3 v = vehicle2->velocity - vehicle1->velocity;
		float a = v*p;
		float b = v.length();

		//std::cout << "    a: " << a << std::endl;
		//std::cout << "    b: " << b << std::endl;

		//test to see what situation we've got
		if (a > -EPS && a < EPS)	
		{
			//std::cout << "    a < " << fabs(EPS) << std::endl;
			return NO_COLLISION;	// vehicle2 is moving parallel to vehicle1
		}
		if (a > EPS)
		{
			//std::cout << "    a > " << EPS << std::endl;
			return NO_COLLISION;	// closest approach was in the past
		}
		if (b > -EPS && b < EPS)
		{
			//std::cout << "    b < " << fabs(EPS) << std::endl;
			return NO_COLLISION;	// no relative motion
		}

		if (a < -EPS && !(b > -EPS && b < EPS))			// we have a closest approach
		{
			//std::cout << "    we have a candidate!" << std::endl;
			time = - a / sqr(b);
			
			if (time < maxTime)						// we need to take action!
			{
				distanceSqr = sqr(p.length()) + 2*a*time + sqr(b)*sqr(time);
				return COLLISION;
			}
		}

		return NO_COLLISION;

		 
	}

	Vec3 calculate() {

		std::vector<B*>::iterator it = allBoids->begin();
		
		float time, distance;
		float closestTime = 99999999999.0f;
		float closestDist = 99999999999.0f;
		B* closestBoid = 0;

		for (; it != allBoids->end(); ++it)
		{
			// dereference the iterator to give the pointer!!
			// mmmmmm i love pointers!
			// if this is not the boid this behaviour is attached to
			// OPT: WE'LL BE CHECKING EACH ONE TWICE SO ADD TIMESTAMP AND ONLY CHECK IF IT HAS NOT BEEN CHECKED RECENTLY
			if (*it != thisBoid)
			{
				//std::cout << "=>Checking boid for collision..." << std::endl;
				if (findClosestApproach(thisBoid, *it, 3, time, distance) == COLLISION)
				{

					//std::cout << "        =>FOUND COLLISION!: " << time << std::endl;
					
					if (time < closestTime)
					{
						
						closestTime = time;
						closestDist = distance;
						closestBoid = *it;
					}
				}
			}
		}

		// check to see if there is a boid we need to worry about!
		if (closestBoid)
		{
			//take evasive action! just turn away from it
			Vec3 meanVelocity = (thisBoid->velocity + closestBoid->velocity) / 2;
			if (fabs(meanVelocity.length()) < EPS) // if they're heading straight at each other
			{
				// just turn drastically, steer sideways!
				return thisBoid->side;// - thisBoid->velocity;
			}
			// otherwise just scale the currentvelocity-meanvelocity
			return thisBoid->velocity - meanVelocity * 500;
		}

		return Vec3(0, 0, 0);
	}

	std::vector<B*>* allBoids;
	B* thisBoid;
};



class BSeperation : public Behaviour
{
public:
	BSeperation(std::vector<B*>* _allBoids, B* _thisBoid, float _weight){
		allBoids = _allBoids;
		thisBoid = _thisBoid;
		weight = _weight;
		bId = BID_SEPERATION;
	}
	~BSeperation(){};

	Vec3 calculate(){

		Vec3 steerFrc;
		std::vector<B*>::iterator it = allBoids->begin();
		for (; it != allBoids->end(); ++it)
		{
			if (*it != thisBoid)
			{
				Vec3 to = (*it)->position - thisBoid->position;
				if (to.length())
				{
					float r = 1.0f / cub(to.length());
							//std::cout << "to: " << to << std::endl;
							//std::cout << "to.l: " << to.length() << std::endl;
							//std::cout << "to.l^2: " << sqr(to.length()) << std::endl;
							//std::cout << "r: " << r << std::endl;
					steerFrc -= to.normal() * r;
				}
			}
			
		}
		
		//std::cout << "    SFrc: " << steerFrc << std::endl;
		return steerFrc;

	}

	std::vector<B*>* allBoids;
	B* thisBoid;
};



class Boid : public B
{
public:
	Boid(){
		top = Vec3(0, 0, 0.65);
		left = Vec3(-0.3, 0, -0.35);
		right = Vec3(0.3, 0, -0.35);

		mass = 1;
		maxForce = 1;
		maxSpeed = 25;

		drawStrFrc = false;
		avoidWalls = 0;
		avoidBoids = 0;
	}
	
	Boid(Vec3 _pos, Vec3 _heading, Vec3 _up){
		position = _pos;
		heading = _heading;
		up = _up;
		side = up ^ heading;

		top = Vec3(0, 0, 0.65);
		left = Vec3(-0.3, 0, -0.35);
		right = Vec3(0.3, 0, -0.35);

		mass = 1;
		maxForce = 1;
		maxSpeed = 25;
		drawStrFrc = false;
		collisionRadius = 0.65f;

		avoidWalls = 0;
		avoidBoids = 0;
	}

	Boid(Vec3 _pos, Vec3 _heading, Vec3 _up, 
			float _mass, float _maxSpeed, float _maxForce){
		position = _pos;
		heading = _heading;
		up = _up;
		side = up ^ heading;

		top = Vec3(0, 0, 0.65);
		left = Vec3(-0.3, 0, -0.35);
		right = Vec3(0.3, 0, -0.35);

		mass = _mass;
		maxForce = _maxForce;
		maxSpeed = _maxSpeed;
		drawStrFrc = false;
		collisionRadius = 0.65f;

		avoidWalls = 0;
		avoidBoids = 0;
	}

	~Boid()
	{
		// we must delete all the behaviour objects we have added to this boid
		for (unsigned i = 0; i < behaviourList.size(); i++)
		{
			delete behaviourList[i];
		}
		if (avoidWalls) delete avoidWalls;
		
		if (avoidBoids)
		{
			std::vector<B*>::iterator it = avoidBoids->allBoids->begin();
			for (; it != avoidBoids->allBoids->end(); ++it)
			{
				if (*it) delete *it;
			}
		}
		
	
	}

	void addBehaviour(Behaviour* b)
	{
		behaviourList.push_back(b);
	}

	inline void updateBasisFromHeading() {
		heading.normalize();
		side = (up ^ heading).normal();
		up = (heading ^ side).normal();

	}

	void draw(int two) {
		Mat4 t;
		
		Vec3 avHeading, avSide, avUp;

		avHeading = ((heading + lastH + twoH + threeH) / 4).normal();

		avSide = (up ^ avHeading).normal();
		avUp = (avHeading ^ avSide).normal();
		
		t._00 = avSide.x;	t._10 = avSide.y;	t._20 = avSide.z;	t._30 = -(position*avSide);
		t._01 = avUp.x;		t._11 = avUp.y;		t._21 = avUp.z;		t._31 = -(position*avUp);
		t._02 = avHeading.x;		t._12 = avHeading.y;		t._22 = avHeading.z;		t._32 = -(position*avHeading);
	

		//printf("t:\n");
		//t.print();

		t = t.inverse();

		//printf("t:\n");
		//t.print();

		Vec3 _top = t.multAsPoint(top);
		Vec3 _left = t.multAsPoint(left);
		Vec3 _right = t.multAsPoint(right);
		
		//printf("Norm:\nl: %f %f %f\nr: %f %f %f\nt:%f %f %f\n\n", left.x, left.y, left.z, right.x, right.y, right.z, top.x, top.y, top.z);
		//printf("Trans:\n_l: %f %f %f\n_r: %f %f %f\n_t:%f %f %f\n\n", _left.x, _left.y, _left.z, _right.x, _right.y, _right.z, _top.x, _top.y, _top.z);
		//fflush(stdout);
		
		if (two)
			glColor3f(colour.r, colour.g, colour.b);
		else
			glColor3f(1, 0.9, 0);
		
		glPointSize(5);
		glBegin(GL_POINTS);/*
		glVertex3f(_left.x, _left.y, _left.z);
		glVertex3f(_right.x, _right.y, _right.z);
		glVertex3f(_top.x, _top.y, _top.z);*/
		glVertex3f(position.x, position.y, position.z);
		glEnd();
		
		if (drawStrFrc)
		{
			glLineWidth(2);
			glColor3f(0, 1, 0);
			glBegin(GL_LINES);
			glVertex3f(position.x, position.y, position.z);
			glVertex3f(position.x + cumStrFrc.x*100,position.y + cumStrFrc.y*100,position.z + cumStrFrc.z*100);
			glEnd();
		}
	}
	
	
	
	// update the velocity and position
	void update(float timeElapsed){
		
		timeElapsed = fabs(timeElapsed);

		cumStrFrc = Vec3(0, 0, 0);
		

		// first check that we don't need to avoid any walls, boid MUST NOT go through walls
		Vec3 avoidWallFrc = avoidWalls->calculate();
		if (avoidWallFrc.length())
		{
			//glColor3f(0, 1, 0);
			//glBegin(GL_LINES);
			//glVertex3f(position.x, position.y, position.z);
			//glVertex3f(position.x + avoidWallFrc.x, position.y + avoidWallFrc.y, position.z + avoidWallFrc.z);
			//glEnd();

			cumStrFrc = avoidWallFrc;
			cumStrFrc.truncate(maxForce);
			Vec3 accel = cumStrFrc / mass;
			velocity += accel * timeElapsed;
			velocity.truncate(maxSpeed);
		}
		else
		{
			// then check that we don't need to avoid any other boids, boids SHOULD NOT go through each other
			Vec3 avoidBoidFrc = avoidBoids->calculate();
			if (avoidBoidFrc.length())
			{
				cumStrFrc = avoidBoidFrc;
				cumStrFrc.truncate(maxForce);
				Vec3 accel = cumStrFrc / mass;
				velocity += accel *timeElapsed;
				velocity.truncate(maxSpeed);
			}
			else
			{
				for (unsigned i = 0; i < behaviourList.size(); i++)
				{
					// weight and sum the resultant force of each behaviour
					cumStrFrc += behaviourList[i]->calculate() * behaviourList[i]->weight;
				}

				cumStrFrc.truncate(maxForce);
				Vec3 accel = cumStrFrc / mass;
				velocity += accel * timeElapsed;
				velocity.truncate(maxSpeed);
			}
			
			
		}
		

		
        

		position += velocity;// * timeElapsed;
		
		threeH = twoH;
		twoH = lastH;
		heading = velocity.normal();
		lastH = heading;
		updateBasisFromHeading();

	}

	std::vector<Behaviour*> behaviourList;
	Vec3 cumStrFrc;
	bool drawStrFrc;
	float collisionRadius;

	// last three heading vectors to average
	Vec3 lastH, twoH, threeH;
	

	// Collision avoidance behaviours that we always want to implement
	BAvoidWalls*	avoidWalls;
	BUCA*			avoidBoids;

};

inline std::ostream& operator << (std::ostream& os, Boid& b)
{
	os << "P: " << b.position << "\nH: " << b.heading << "\nU: " << b.up << "\nS: " << b.side;
	return os;
}